BookFusion.Audio = (() => {
  let audioElementIds = []
  let mediaActiveClass = '-epub-media-overlay-active'

  function isInside (element, rect) {
    const elementRect = BookFusion.Common.getElementRect(element)
    return elementRect.y >= rect.y &&
      elementRect.y < rect.y + rect.height &&
      elementRect.x >= rect.x &&
      elementRect.x < rect.x + rect.width
  }

  // now it returns same element if it is playable
  function getClosestAudioElement (elementId, rawRect) {
    if (!audioElementIds.includes(elementId)) {
      return
    }

    var rect = BookFusion.Transformation.transformRectFromNative(rawRect)

    const audioElement = document.getElementById(elementId)
    if (audioElement && isInside(audioElement, rect)) {
      return elementId
    }
  }

  function getFirstAudioElementIdInRect (rawRect) {
    for (const id of audioElementIds) {
      const audioElement = document.getElementById(id)
      if (!audioElement) {
        continue
      }

      var rect = BookFusion.Transformation.transformRectFromNative(rawRect)

      if (isInside(audioElement, rect)) {
        return id
      }
    }
  }

  function getFirstAudioElementIdByOffsets (offset) {
    const range = BookFusion.RangeOffsetUtils.createRangeByOffsets(offset, offset + 1)

    if (!range) {
      return
    }

    return getAudioElementIdContainingNode(range.startContainer)
  }

  function getAudioElementIdContainingNode (node) {
    if (node.nodeType === Node.ELEMENT_NODE && typeof node.id !== 'undefined' && audioElementIds.includes(node.id)) {
      return node.id
    }

    if (node.parentElement) {
      return getAudioElementIdContainingNode(node.parentElement)
    }
  }

  function setAudioElementIds (ids) {
    audioElementIds = ids
  }

  function setMediaActiveClass (className) {
    mediaActiveClass = className
  }

  function highlightActiveAudio (elementId) {
    unhighlightActiveAudio()

    const element = document.getElementById(elementId)
    if (!element) {
      return
    }
    element.classList.add(mediaActiveClass)
  }

  function unhighlightActiveAudio () {
    for (const activeAudio of document.querySelectorAll(`.${mediaActiveClass}`)) {
      activeAudio.classList.remove(mediaActiveClass)
    }

    unhighlightSpeechPosition()

    BookFusion.Highlights.clearSelection()
  }

  function findStartEndOffset (rawRect) {
    const isHorizontal = !BookFusion.Common.isScrollModeEnabled()

    var rect = BookFusion.Transformation.transformRectFromNative(rawRect)

    const xWithShift = rect.x - window.scrollX
    const yWithShift = rect.y - window.scrollY

    const [start, end] = isHorizontal
      ? [xWithShift, xWithShift + rect.width]
      : [yWithShift, yWithShift + rect.height]

    const startRange = FuturePress.EPUB.findStart(
      document.body,
      start,
      end,
      isHorizontal,
      BookFusion.Common.isRTL()
    )

    if (!startRange) {
      return
    }

    const endRange = FuturePress.EPUB.findEnd(
      document.body,
      start,
      end,
      isHorizontal,
      BookFusion.Common.isRTL()
    )

    if (!endRange) {
      return
    }

    const startOffsets = BookFusion.RangeOffsetUtils.findOffsetsByRange(startRange)
    if (!startOffsets) {
      return
    }

    const endOffsets = BookFusion.RangeOffsetUtils.findOffsetsByRange(endRange)
    if (!endOffsets) {
      return
    }

    const range = BookFusion.RangeOffsetUtils.createRangeByOffsets(
        startOffsets.startOffset,
        endOffsets.endOffset
    )

    return {
        text: range.toString(),
        startOffset: startOffsets.startOffset,
        endOffset: endOffsets.endOffset
    }
  }

  function getTextStartedAtOffset (startOffset, endOffset) {
    if (startOffset === endOffset) {
      return
    }

    const range = BookFusion.RangeOffsetUtils.createRangeByOffsets(startOffset, endOffset)

    if (!range) {
      return
    }

    const textFromRect = range.toString()

    if (!textFromRect || textFromRect.trim().length === 0) {
      return
    }

    range.selectNode(document.body)

    return { currentOffset: startOffset, text: range.toString() }
  }

  function getTextInRect (rect, offset) {
    const { startOffset, endOffset } = findStartEndOffset(BookFusion.Transformation.convertToWebViewFrame(rect))
    if (offset && offset > startOffset && offset < endOffset) {
      return getTextStartedAtOffset(offset, endOffset) || getTextStartedAtOffset(startOffset, endOffset)
    }
    return getTextStartedAtOffset(startOffset, endOffset)
  }

  function getAudioDetails (rawRect, offset) {
    var rect = BookFusion.Transformation.transformRectFromNative(rawRect)
    const mediaOverlay = getFirstAudioElementIdInRect(rect)
    if (mediaOverlay) {
      return { mediaOverlay }
    }
    return getTextInRect(rect, offset)
  }

  const speechPositionHighlightId = 'speechPositionHighlightId'

  function highlightSpeechPosition (startOffset, endOffset, style) {
    BookFusion.Highlights.addHighlight(startOffset, endOffset, style, speechPositionHighlightId)
  }

  function unhighlightSpeechPosition () {
    BookFusion.Highlights.removeHighlight(speechPositionHighlightId)
  }

  return {
    isInside,
    setAudioElementIds,
    getFirstAudioElementIdInRect,
    getFirstAudioElementIdByOffsets,
    getClosestAudioElement,
    setMediaActiveClass,
    highlightActiveAudio,
    highlightSpeechPosition,
    unhighlightActiveAudio,
    getAudioDetails,
    findStartEndOffset,
    getAudioElementIdContainingNode,
  }
})()