// https://github.com/futurepress/epub.js/blob/master/src/mapping.js

const FuturePress = {}

FuturePress.EPUB = (() => {
  function walk (root, func) {
    // IE11 has strange issue, if root is text node IE throws exception on
    // calling treeWalker.nextNode(), saying
    // Unexpected call to method or property access instead of returning null value
    if (root && root.nodeType === Node.TEXT_NODE) {
      return
    }

    const nodeFilter = {
      acceptNode: function (node) {
        if (node.data.trim().length > 0) {
          return NodeFilter.FILTER_ACCEPT
        } else {
          return NodeFilter.FILTER_REJECT
        }
      }
    }

    const treeWalker = document.createTreeWalker(root, NodeFilter.SHOW_TEXT, nodeFilter, false)
    let node
    while ((node = treeWalker.nextNode())) {
      const result = func(node)
      if (result) {
        return result
      }
    }
  }

  function splitTextNodeIntoRanges (node, _separator) {
    const text = (node.textContent || '').trim()
    const document = node.ownerDocument
    const separator = _separator || ' '

    let separatorPosition = text.indexOf(separator)

    if (separatorPosition === -1 || node.nodeType !== Node.TEXT_NODE) {
      const range = document.createRange()
      range.selectNodeContents(node)
      return [range]
    }

    let range = document.createRange()
    range.setStart(node, 0)
    range.setEnd(node, separatorPosition)
    const ranges = [range]
    range = false

    while (separatorPosition !== -1) {
      separatorPosition = text.indexOf(separator, separatorPosition + 1)
      if (separatorPosition > 0) {
        if (range) {
          range.setEnd(node, separatorPosition)
          ranges.push(range)
        }
        range = document.createRange()
        range.setStart(node, separatorPosition + 1)
      }
    }

    if (range) {
      range.setEnd(node, text.length)
      ranges.push(range)
    }

    return ranges
  }

  function nodeBounds (node) {
    const document = node.ownerDocument
    if (node.nodeType === Node.TEXT_NODE) {
      const range = document.createRange()
      range.selectNodeContents(node)
      return range.getBoundingClientRect()
    } else {
      return node.getBoundingClientRect()
    }
  }

  function findTextStartRange (node, start, end, isHorizontal, isRTL) {
    const ranges = splitTextNodeIntoRanges(node)

    for (let i = 0; i < ranges.length; i++) {
      const range = ranges[i]

      const bounds = range.getBoundingClientRect()

      if (isHorizontal && !isRTL) {
        if (bounds.left >= start) {
          return range
        }
      } else if (isHorizontal && isRTL) {
        if (bounds.right <= end) {
          return range
        }
      } else {
        if (bounds.top >= start) {
          return range
        }
      }
    }

    return ranges[0]
  }

  function findTextEndRange (node, start, end, isHorizontal, isRTL) {
    const ranges = splitTextNodeIntoRanges(node)
    let previousRange

    for (let i = 0; i < ranges.length; i++) {
      const range = ranges[i]

      const bounds = range.getBoundingClientRect()

      if (isHorizontal && !isRTL) {
        const left = bounds.left
        const right = bounds.right

        if (left > end && previousRange) {
          return previousRange
        } else if (right > end) {
          return range
        }
      } else if (isHorizontal && isRTL) {
        const left = bounds.left
        const right = bounds.right

        if (right < start && previousRange) {
          return previousRange
        } else if (left < start) {
          return range
        }
      } else {
        const top = bounds.top
        const bottom = bounds.bottom
        if (top > end && previousRange) {
          return previousRange
        } else if (bottom > end) {
          return range
        }
      }

      previousRange = range
    }

    // Ends before limit
    return ranges[ranges.length - 1]
  }

  function findStart (rootElement, start, end, isHorizontal, isRTL) {
    const stack = [rootElement]
    let previousElement = rootElement

    while (stack.length) {
      const currentElement = stack.shift()

      const foundElement = walk(currentElement, (node) => {
        const bounds = nodeBounds(node)

        if (isHorizontal && !isRTL) {
          const left = bounds.left
          const right = bounds.right

          if (left >= start && left <= end) {
            return node
          } else if (right > start) {
            return node
          } else {
            previousElement = node
            stack.push(node)
          }
        } else if (isHorizontal && isRTL) {
          const left = bounds.left
          const right = bounds.right

          if (right <= end && right >= start) {
            return node
          } else if (left < end) {
            return node
          } else {
            previousElement = node
            stack.push(node)
          }
        } else {
          const top = bounds.top
          const bottom = bounds.bottom

          if (top >= start && top <= end) {
            return node
          } else if (bottom > start) {
            return node
          } else {
            previousElement = node
            stack.push(node)
          }
        }
      })

      if (foundElement) {
        return findTextStartRange(foundElement, start, end, isHorizontal, isRTL)
      }
    }

    // Return last element
    return findTextStartRange(previousElement, start, end, isHorizontal, isRTL)
  }

  function findEnd (rootElement, start, end, isHorizontal, isRTL) {
    const stack = [rootElement]
    let previousElement = rootElement

    while (stack.length) {
      const currentElement = stack.shift()

      const foundElement = walk(currentElement, (node) => {
        const bounds = nodeBounds(node)

        if (isHorizontal && !isRTL) {
          const left = Math.round(bounds.left)
          const right = Math.round(bounds.right)

          if (left > end && previousElement) {
            return previousElement
          } else if (right > end) {
            return node
          } else {
            previousElement = node
            stack.push(node)
          }
        } else if (isHorizontal && isRTL) {
          const left = Math.round(bounds.left)
          const right = Math.round(bounds.right)

          if (right < start && previousElement) {
            return previousElement
          } else if (left < start) {
            return node
          } else {
            previousElement = node
            stack.push(node)
          }
        } else {
          const top = Math.round(bounds.top)
          const bottom = Math.round(bounds.bottom)

          if (top > end && previousElement) {
            return previousElement
          } else if (bottom > end) {
            return node
          } else {
            previousElement = node
            stack.push(node)
          }
        }
      })

      if (foundElement) {
        return findTextEndRange(foundElement, start, end, isHorizontal, isRTL)
      }
    }

    // end of chapter
    return findTextEndRange(previousElement, start, end, isHorizontal, isRTL)
  }

  return {
    findStart,
    findEnd
  }
})()