BookFusion.Highlights = (() => {
  document.addEventListener('selectionchange', (event) => {
    const selection = window.getSelection()
    if (selection.rangeCount === 0) {
      window.BookFusionAppBridge.onSelectedTextFinished()
      return
    }

    const range = selection.getRangeAt(0)
    const rect = BookFusion.Common.getRangeRect(range)
    const offsets = BookFusion.RangeOffsetUtils.findOffsetsByRange(range)
    if (offsets.startOffset === offsets.endOffset) {
      window.BookFusionAppBridge.onSelectedTextFinished()
      return
    }

    // during creation of highlight we send text as is (without removing unneeded spaces)
    const text = range.toString()
    const textHtml = rangeToHTML(range)
    const message = Object.assign({ text, textHtml }, rect, offsets)
    window.BookFusionAppBridge.onSelectedTextChanged(JSON.stringify(message))
  })

  let highlights = []

  function highlight (localId, startOffset, endOffset, className, style, onClick) {
    const range = BookFusion.RangeOffsetUtils.createRangeByOffsets(startOffset, endOffset)

    if (!range) {
      BookFusion.Common.reportError(`Invalid range ${startOffset}-${endOffset}`)
      return
    }

    const nodes = []

    if (range.startContainer === range.endContainer) {
      nodes.push(range.startContainer)
    } else {
      const treeWalker = document.createTreeWalker(
        range.commonAncestorContainer,
        NodeFilter.SHOW_TEXT,
        function (node) {
          return BookFusion.RangeOffsetUtils.rangeIntersectsNode(range, node)
            ? NodeFilter.FILTER_ACCEPT
            : NodeFilter.FILTER_REJECT
        },
        false
      )

      while (treeWalker.nextNode()) {
        nodes.push(treeWalker.currentNode)
      }
    }

    const highlight = {
      localId,
      text: range.toString().replace(/\n/g, ' '),
      textHtml: rangeToHTML(range),
      startOffset,
      endOffset,
      wrappers: []
    }

    nodes.forEach((node) => {
      if (isBlank(node.textContent)) {
        return
      }

      const replaceRange = document.createRange()
      replaceRange.selectNodeContents(node)
      if (node === range.startContainer) {
        replaceRange.setStart(range.startContainer, range.startOffset)
      }
      if (node === range.endContainer) {
        replaceRange.setEnd(range.endContainer, range.endOffset)
      }

      const fragment = replaceRange.extractContents()
      const wrapper = document.createElement('span')
      wrapper.className = className
      wrapper.appendChild(fragment)
      if (onClick) {
        wrapper.addEventListener('click', onClick(highlight))
      }
      setElementStyle(wrapper, style)
      replaceRange.insertNode(wrapper)

      highlight.wrappers.push(wrapper)
    })

    return highlight
  }

  function _addHighlight (localId, startOffset, endOffset, style) {
    const onClick = highlight => {
      return event => {
        const point = BookFusion.Common.convertToWebViewPoint({ x: event.clientX, y: event.clientY })
        const message = Object.assign({
          localId: highlight.localId,
          text: highlight.text,
          textHtml: highlight.textHtml,
          startOffset,
          endOffset
        }, point)
        window.BookFusionAppBridge.onHighlightClicked(JSON.stringify(message))
      }
    }

    const newHighlight = highlight(localId, startOffset, endOffset, 'bf-highlight', style, onClick)

    highlights.push(newHighlight)
  }

  function setHighlights (highlights) {
    removeAllHighlights()
    highlights.forEach(highlight =>
      _addHighlight(highlight.localId, highlight.textRange.startOffset, highlight.textRange.endOffset, highlight.style)
    )
  }

  function setElementStyle (element, style) {
    element.style.backgroundColor = style && style.backgroundColor
    element.style.textDecorationStyle = style && style.textDecorationStyle
    element.style.textDecorationColor = style && style.textDecorationColor
    element.style.textDecorationLine = style && style.textDecorationLine
  }

  function addHighlight (highlight) {
    _addHighlight(highlight.localId, highlight.textRange.startOffset, highlight.textRange.endOffset, highlight.style)
    clearSelection()
  }

  function setHighlightStyle ({ localId, startOffset, endOffset }, style) {
    const index = highlights.findIndex(highlight =>
      highlight.localId === localId
    )

    if (index === -1) {
      return
    }

    const highlight = highlights[index]
    highlight.wrappers.forEach(wrapper =>
      setElementStyle(wrapper, style)
    )
  }

  function removeHighlight ({ localId, startOffset, endOffset }) {
    const index = highlights.findIndex(highlight =>
      highlight.localId === localId
    )

    if (index === -1) {
      return
    }

    const highlight = highlights[index]
    unhighlight(highlight)

    highlights.splice(index, 1)
  }

  function removeAllHighlights () {
    highlights.forEach(highlight => {
      unhighlight(highlight)
    })
    highlights = []
  }

  function clearSelection () {
    window.getSelection().removeAllRanges()
  }

  function unhighlight ({ wrappers }) {
    wrappers.forEach(wrapper => {
      const parent = wrapper.parentNode
      const children = Array.prototype.slice.call(wrapper.childNodes)
      children.forEach(child => {
        parent.insertBefore(child, wrapper)
      })
      wrapper.remove()
      parent.normalize()
    })
  }

  function isBlank (string) {
    return !string || !!string.match(/^\s*$/)
  }

  function rangeToHTML (range) {
    const element = document.createElement('div')
    element.appendChild(range.cloneContents())
    return element.innerHTML
  }

  return {
    setHighlights,
    addHighlight,
    setHighlightStyle,
    removeHighlight,
    highlight,
    unhighlight,
    clearSelection
  }
})()