BookFusion.TapHandler = (() => {
  const imageDoubleClick = {
    isTriggered: false,
    duration: 200
  }

  const imageLongTouch = {
    isTriggered: false,
    duration: 800,
    timer: null,
    startPosition: null,
    allowedMove: 10
  }

  window.addEventListener('DOMContentLoaded', () => {
    window.addEventListener('touchstart',
      event => {
        imageLongTouch.isTriggered = false
        imageLongTouch.startPosition = getTouchEventPosition(event)

        event.preventDefault()
        const target = event.target
        if (!imageLongTouch.timer && isImageElement(target)) {
          startLongTouchTimer(target)
        }
      },
      false
    )

    window.addEventListener('touchend', cancelLongTouchHandler, false)
    window.addEventListener('touchcancel', cancelLongTouchHandler, false)

    window.addEventListener('touchmove', event => {
      const newPosition = getTouchEventPosition(event)
      if (imageLongTouch.startPosition && newPosition && getFastDistanceBetweenPoints(imageLongTouch.startPosition, newPosition) < imageLongTouch.allowedMove) {
        return
      }
      cancelLongTouchHandler()
    }, false)

    document.documentElement.onclick = event => {
      if (isButton(event.target)) {
        return
      }

      if (isImageElement(event.target)) {
        if (imageLongTouch.isTriggered || imageDoubleClick.isTriggered) {
          return
        }

        setTimeout(() => {
          if (!imageDoubleClick.isTriggered) {
            postClickMessage(event)
          }
        }, imageDoubleClick.duration)
      } else {
        // Android: special case for link handling.
        const linkHref = getLinkHref(event.target)
        if (linkHref != null) {
          event.preventDefault()
          window.BookFusionAppBridge.onLinkClicked(linkHref)
        } else if (isStaticElement(event.target)) {
          postClickMessage(event)
        }
      }
    }

    document.documentElement.ondblclick = event => {
      const target = event.target
      if (!isImageElement(target)) {
        return
      }

      imageDoubleClick.isTriggered = true
      setTimeout(() => { imageDoubleClick.isTriggered = false }, imageDoubleClick.duration)

      const imageSource = getImageElementSource(target)
      if (imageSource) {
        window.BookFusionAppBridge.onImageClicked(imageSource)
      }
    }
  })

  function postClickMessage (event) {
    var pixelRatio = window.devicePixelRatio;
    const coordinates = {
      x: event.pageX * pixelRatio,
      y: event.pageY * pixelRatio
    }
    window.BookFusionAppBridge.onContentClicked(JSON.stringify(coordinates))
  }

  function getUppercasedTagName (element) {
    return element.tagName.toUpperCase()
  }

  function getImageElementSource (element) {
    const tagName = getUppercasedTagName(element)
    if (tagName === 'IMG') {
      return element.src
    } else if (tagName === 'IMAGE') {
      // https://developer.mozilla.org/en-US/docs/Web/SVG/Element/image
      const href = element.getAttribute('xlink:href') || element.getAttribute('href')
      if (href) {
        // resolve possible relative path
        const tempLink = document.createElement('a');
        tempLink.href = href
        return tempLink.href
      }
    }
  }

  function isImageElement (element) {
    const tagName = getUppercasedTagName(element)
    return tagName === 'IMG' || tagName === 'IMAGE'
  }

  function isStaticElement (element) {
    const tagName = getUppercasedTagName(element)

    if (tagName === 'AUDIO') {
      return false
    } else if (tagName === 'VIDEO') {
      return false
    } else if (element.classList.contains('bf-highlight')) {
      return false
    } else if (isAnchor(element)) {
      return false
    }

    return true
  }

  function isAnchor (element) {
    if (getUppercasedTagName(element) === 'A') {
      return !!element.href || !!element.onclick
    } else if (element.parentElement) {
      return isAnchor(element.parentElement)
    } else {
      return false
    }
  }

  function isButton (element) {
    if (getUppercasedTagName(element) === 'BUTTON' || element.role === 'button' || element.closest('[role=button]')) {
      return true
    } else {
      return false
    }
  }

  function getTouchEventPosition (event) {
    const touch = event.touches[0] || event.changedTouches[0]
    if (!touch || typeof touch.clientX === 'undefined' || typeof touch.clientY === 'undefined') {
      return
    }

    return { x: touch.clientX, y: touch.clientY }
  }

  function getFastDistanceBetweenPoints (point1, point2) {
    return Math.abs(point2.x - point1.x) + Math.abs(point2.y - point1.y)
  }

  function handleLongTouchForImage (image) {
    const message = getSelectedImageMessageForElement(image)
    window.webkit.messageHandlers.didSelectImage.postMessage(message)
  }

  function startLongTouchTimer (image) {
    cancelLongTouchHandler()

    imageLongTouch.timer = setTimeout(() => {
      imageLongTouch.timer = null
      imageLongTouch.isTriggered = true
      handleLongTouchForImage(image)
    }, imageLongTouch.duration)
  }

  function cancelLongTouchHandler () {
    if (imageLongTouch.timer) {
      clearTimeout(imageLongTouch.timer)
      imageLongTouch.timer = null
    }
  }

  function getSelectedImageMessageForElement (image) {
    const offset = BookFusion.RangeOffsetUtils.getElementOffset(image)
    const { left, top, width, height } = FuturePress.CFI.getContentImageRect(image, 'contain')
    const cfi = BookFusion.CFI.getCFI(image).toString()
    return { x: left, y: top, width, height, cfi, offset }
  }

  // Android only
  function getLinkHref (element) {
    if (element.tagName.toUpperCase() === 'A') {
      return element.href
    } else if (element.parentElement) {
      return getLinkHref(element.parentElement)
    } else {
      return null
    }
  }
})()